@@ -25,14 +25,7 @@
 #endif
 
 #include <string.h>
-#include <unistd.h>
-#include <errno.h>
-#include <sys/socket.h>
-#include <linux/if.h>
-#include <linux/if_packet.h>
 #include <linux/if_ether.h>
-#include <arpa/inet.h>
-#include <linux/filter.h>
 #include <ell/ell.h>
 
 #include "crypto.h"
@@ -47,133 +40,15 @@
 struct l_queue *state_machines;
 struct l_queue *preauths;
 struct watchlist frame_watches;
+static uint32_t eapol_4way_handshake_time = 2;
 
-eapol_deauthenticate_func_t deauthenticate = NULL;
 eapol_rekey_offload_func_t rekey_offload = NULL;
 
-static struct l_io *pae_io;
 eapol_tx_packet_func_t tx_packet = NULL;
 void *tx_user_data;
 
 uint32_t next_frame_watch_id;
 
-/*
- * BPF filter to match skb->dev->type == 1 (ARPHRD_ETHER) and
- * match skb->protocol == 0x888e (PAE) or 0x88c7 (preauthentication).
- */
-static struct sock_filter pae_filter[] = {
-	{ 0x28,  0,  0, 0xfffff01c },	/* ldh #hatype		*/
-	{ 0x15,  0,  4, 0x00000001 },	/* jne #1, drop		*/
-	{ 0x28,  0,  0, 0xfffff000 },	/* ldh #proto		*/
-	{ 0x15,  1,  0, 0x0000888e },	/* je  #0x888e, keep	*/
-	{ 0x15,  0,  1, 0x000088c7 },	/* jne #0x88c7, drop	*/
-	{ 0x06,  0,  0, 0xffffffff },	/* keep: ret #-1	*/
-	{ 0x06,  0,  0, 0000000000 },	/* drop: ret #0		*/
-};
-
-static const struct sock_fprog pae_fprog = { .len = 7, .filter = pae_filter };
-
-static struct l_io *pae_open(void)
-{
-	struct l_io *io;
-	int fd;
-
-	fd = socket(PF_PACKET, SOCK_DGRAM | SOCK_CLOEXEC | SOCK_NONBLOCK,
-							htons(ETH_P_ALL));
-	if (fd < 0)
-		return NULL;
-
-	if (setsockopt(fd, SOL_SOCKET, SO_ATTACH_FILTER,
-					&pae_fprog, sizeof(pae_fprog)) < 0) {
-		close(fd);
-		return NULL;
-	}
-
-	io = l_io_new(fd);
-	l_io_set_close_on_destroy(io, true);
-
-	return io;
-}
-
-static bool pae_read(struct l_io *io, void *user_data)
-{
-	int fd = l_io_get_fd(io);
-	struct sockaddr_ll sll;
-	socklen_t sll_len;
-	ssize_t bytes;
-	uint8_t frame[IEEE80211_MAX_DATA_LEN];
-
-	memset(&sll, 0, sizeof(sll));
-	sll_len = sizeof(sll);
-
-	bytes = recvfrom(fd, frame, sizeof(frame), 0,
-				(struct sockaddr *) &sll, &sll_len);
-	if (bytes <= 0) {
-		l_error("EAPoL read socket: %s", strerror(errno));
-		return false;
-	}
-
-	if (sll.sll_halen != ETH_ALEN)
-		return true;
-
-	__eapol_rx_packet(sll.sll_ifindex, sll.sll_addr,
-				ntohs(sll.sll_protocol), frame, bytes);
-
-	return true;
-}
-
-static void pae_destroy()
-{
-	pae_io = NULL;
-}
-
-static void pae_write(uint32_t ifindex, const uint8_t *aa, const uint8_t *spa,
-			uint16_t proto, const struct eapol_frame *ef)
-{
-	size_t frame_size;
-	struct sockaddr_ll sll;
-	ssize_t r;
-	int fd;
-
-	if (!pae_io) {
-		if (tx_packet) /* Used for unit tests */
-			tx_packet(ifindex, aa, spa, ef, tx_user_data);
-
-		return;
-	}
-
-	fd = l_io_get_fd(pae_io);
-
-	memset(&sll, 0, sizeof(sll));
-	sll.sll_family = AF_PACKET;
-	sll.sll_ifindex = ifindex;
-	sll.sll_protocol = htons(proto);
-	sll.sll_halen = ETH_ALEN;
-	memcpy(sll.sll_addr, aa, ETH_ALEN);
-
-	frame_size = sizeof(struct eapol_header) +
-			L_BE16_TO_CPU(ef->header.packet_len);
-
-	r = sendto(fd, ef, frame_size, 0,
-			(struct sockaddr *) &sll, sizeof(sll));
-	if (r < 0)
-		l_error("EAPoL write socket: %s", strerror(errno));
-}
-
-void eapol_pae_open()
-{
-	pae_io = pae_open();
-	if (!pae_io)
-		return;
-
-	l_io_set_read_handler(pae_io, pae_read, NULL, pae_destroy);
-}
-
-void eapol_pae_close()
-{
-	l_io_destroy(pae_io);
-}
-
 #define VERIFY_IS_ZERO(field)						\
 	do {								\
 		if (!util_mem_is_zero((field), sizeof((field))))	\
@@ -269,6 +144,9 @@ uint8_t *eapol_decrypt_key_data(const uint8_t *kek,
 		break;
 	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES:
 	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
+		if (key_data_len < 24 || key_data_len % 8)
+			return NULL;
+
 		expected_len = key_data_len - 8;
 		break;
 	default:
@@ -296,9 +174,6 @@ uint8_t *eapol_decrypt_key_data(const uint8_t *kek,
 	}
 	case EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES:
 	case EAPOL_KEY_DESCRIPTOR_VERSION_AES_128_CMAC_AES:
-		if (key_data_len < 24 || key_data_len % 8)
-			goto error;
-
 		if (!aes_unwrap(kek, key_data, key_data_len, buf))
 			goto error;
 
@@ -447,7 +322,6 @@ bool eapol_verify_ptk_1_of_4(const struct eapol_key *ek)
 	if (ek->wpa_key_id)
 		return false;
 
-	VERIFY_IS_ZERO(ek->eapol_key_iv);
 	VERIFY_IS_ZERO(ek->key_rsc);
 	VERIFY_IS_ZERO(ek->reserved);
 	VERIFY_IS_ZERO(ek->key_mic_data);
@@ -480,6 +354,9 @@ bool eapol_verify_ptk_2_of_4(const struct eapol_key *ek)
 	if (ek->wpa_key_id)
 		return false;
 
+	if (ek->request)
+		return false;
+
 	key_len = L_BE16_TO_CPU(ek->key_length);
 	if (key_len != 0)
 		return false;
@@ -529,12 +406,6 @@ bool eapol_verify_ptk_3_of_4(const struct eapol_key *ek, bool is_wpa)
 
 	VERIFY_IS_ZERO(ek->reserved);
 
-	/* 0 (Version 2) or random (Version 1) */
-	if (ek->key_descriptor_version ==
-			EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES)
-		L_WARN_ON(!util_mem_is_zero(ek->eapol_key_iv,
-						sizeof(ek->eapol_key_iv)));
-
 	return true;
 }
 
@@ -563,6 +434,9 @@ bool eapol_verify_ptk_4_of_4(const struct eapol_key *ek, bool is_wpa)
 	if (ek->wpa_key_id)
 		return false;
 
+	if (ek->request)
+		return false;
+
 	key_len = L_BE16_TO_CPU(ek->key_length);
 	if (key_len != 0)
 		return false;
@@ -619,11 +493,6 @@ bool eapol_verify_gtk_1_of_2(const struct eapol_key *ek, bool is_wpa)
 
 	VERIFY_IS_ZERO(ek->reserved);
 
-	/* 0 (Version 2) or random (Version 1) */
-	if (ek->key_descriptor_version ==
-			EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES)
-		VERIFY_IS_ZERO(ek->eapol_key_iv);
-
 	/*
 	 * WPA_80211_v3_1, Section 2.2.4:
 	 * "Key Index (bits 4 and 5): specifies the key id of the temporal
@@ -780,11 +649,14 @@ struct eapol_sm {
 	void *user_data;
 	struct l_timeout *timeout;
 	struct l_timeout *eapol_start_timeout;
+	unsigned int frame_retry;
+	uint16_t listen_interval;
 	bool have_replay:1;
 	bool started:1;
 	bool use_eapol_start:1;
 	bool require_handshake:1;
 	bool eap_exchanged:1;
+	bool authenticator:1;
 	struct eap_state *eap;
 	struct eapol_frame *early_frame;
 	uint32_t watch_id;
@@ -845,6 +717,11 @@ void eapol_sm_set_protocol_version(struct eapol_sm *sm,
 	sm->protocol_version = protocol_version;
 }
 
+void eapol_sm_set_listen_interval(struct eapol_sm *sm, uint16_t interval)
+{
+	sm->listen_interval = interval;
+}
+
 void eapol_sm_set_user_data(struct eapol_sm *sm, void *user_data)
 {
 	sm->user_data = user_data;
@@ -855,12 +732,16 @@ void eapol_sm_set_event_func(struct eapol_sm *sm, eapol_sm_event_func_t func)
 	sm->event_func = func;
 }
 
+static void eapol_sm_write(struct eapol_sm *sm, const struct eapol_frame *ef,
+				bool noencrypt)
+{
+	__eapol_tx_packet(sm->handshake->ifindex, sm->handshake->aa, ETH_P_PAE,
+				ef, noencrypt);
+}
+
 static inline void handshake_failed(struct eapol_sm *sm, uint16_t reason_code)
 {
-	if (deauthenticate)
-		deauthenticate(sm->handshake->ifindex,
-				sm->handshake->aa, sm->handshake->spa,
-				reason_code, sm->user_data);
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_FAILED, &reason_code);
 
 	eapol_sm_free(sm);
 }
@@ -875,12 +756,6 @@ static void eapol_timeout(struct l_timeout *timeout, void *user_data)
 	handshake_failed(sm, MMPDU_REASON_CODE_4WAY_HANDSHAKE_TIMEOUT);
 }
 
-static void eapol_write(struct eapol_sm *sm, const struct eapol_frame *ef)
-{
-	pae_write(sm->handshake->ifindex,
-			sm->handshake->aa, sm->handshake->spa, ETH_P_PAE, ef);
-}
-
 static void eapol_install_gtk(struct eapol_sm *sm, uint8_t gtk_key_index,
 					const uint8_t *gtk, size_t gtk_len,
 					const uint8_t *rsc)
@@ -922,6 +797,8 @@ static void send_eapol_start(struct l_timeout *timeout, void *user_data)
 	uint8_t buf[sizeof(struct eapol_frame)];
 	struct eapol_frame *frame = (struct eapol_frame *) buf;
 
+	handshake_event(sm->handshake, HANDSHAKE_EVENT_STARTED, NULL);
+
 	l_timeout_remove(sm->eapol_start_timeout);
 	sm->eapol_start_timeout = NULL;
 
@@ -932,7 +809,98 @@ static void send_eapol_start(struct l_timeout *timeout, void *user_data)
 	frame->header.packet_type = 1;
 	l_put_be16(0, &frame->header.packet_len);
 
-	eapol_write(sm, frame);
+	eapol_sm_write(sm, frame, false);
+}
+
+static void eapol_set_key_timeout(struct eapol_sm *sm,
+					l_timeout_notify_cb_t cb)
+{
+	/*
+	 * 802.11-2016 12.7.6.6: "The retransmit timeout value shall be
+	 * 100 ms for the first timeout, half the listen interval for the
+	 * second timeout, and the listen interval for subsequent timeouts.
+	 * If there is no listen interval or the listen interval is zero,
+	 * then 100 ms shall be used for all timeout values."
+	 */
+	unsigned int timeout_ms = 100;
+	unsigned int beacon_us = 100 * 1024;
+
+	sm->frame_retry++;
+
+	if (sm->frame_retry == 2 &&
+			sm->listen_interval != 0)
+		timeout_ms = sm->listen_interval * beacon_us / 2000;
+	else if (sm->frame_retry > 2 &&
+			sm->listen_interval != 0)
+		timeout_ms = sm->listen_interval * beacon_us / 1000;
+
+	if (sm->frame_retry > 1)
+		l_timeout_modify_ms(sm->timeout, timeout_ms);
+	else {
+		if (sm->timeout)
+			l_timeout_remove(sm->timeout);
+
+		sm->timeout = l_timeout_create_ms(timeout_ms, cb, sm,
+								NULL);
+	}
+}
+
+/* 802.11-2016 Section 12.7.6.2 */
+static void eapol_send_ptk_1_of_4(struct eapol_sm *sm)
+{
+	uint32_t ifindex = sm->handshake->ifindex;
+	const uint8_t *aa = sm->handshake->aa;
+	uint8_t frame_buf[512];
+	struct eapol_key *ek = (struct eapol_key *) frame_buf;
+	enum crypto_cipher cipher = ie_rsn_cipher_suite_to_cipher(
+				sm->handshake->pairwise_cipher);
+	uint8_t pmkid[16];
+
+	handshake_state_new_anonce(sm->handshake);
+
+	sm->handshake->ptk_complete = false;
+
+	sm->replay_counter++;
+
+	memset(ek, 0, sizeof(struct eapol_key));
+	ek->header.protocol_version = sm->protocol_version;
+	ek->header.packet_type = 0x3;
+	ek->descriptor_type = EAPOL_DESCRIPTOR_TYPE_80211;
+	/* Must be HMAC-SHA1-128 + AES when using CCMP with PSK or 8021X */
+	ek->key_descriptor_version = EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES;
+	ek->key_type = true;
+	ek->key_ack = true;
+	ek->key_length = L_CPU_TO_BE16(crypto_cipher_key_len(cipher));
+	ek->key_replay_counter = L_CPU_TO_BE64(sm->replay_counter);
+	memcpy(ek->key_nonce, sm->handshake->anonce, sizeof(ek->key_nonce));
+
+	/* Write the PMKID KDE into Key Data field unencrypted */
+	crypto_derive_pmkid(sm->handshake->pmk, sm->handshake->spa, aa,
+			pmkid, false);
+	eapol_key_data_append(ek, HANDSHAKE_KDE_PMKID, pmkid, 16);
+
+	ek->header.packet_len = L_CPU_TO_BE16(sizeof(struct eapol_key) +
+					L_BE16_TO_CPU(ek->key_data_len) - 4);
+
+	l_debug("STA: "MAC" retries=%u", MAC_STR(sm->handshake->spa),
+			sm->frame_retry);
+
+	__eapol_tx_packet(ifindex, sm->handshake->spa, ETH_P_PAE,
+				(struct eapol_frame *) ek, false);
+}
+
+static void eapol_ptk_1_of_4_retry(struct l_timeout *timeout, void *user_data)
+{
+	struct eapol_sm *sm = user_data;
+
+	if (sm->frame_retry >= 3) {
+		handshake_failed(sm, MMPDU_REASON_CODE_4WAY_HANDSHAKE_TIMEOUT);
+		return;
+	}
+
+	eapol_send_ptk_1_of_4(sm);
+
+	eapol_set_key_timeout(sm, eapol_ptk_1_of_4_retry);
 }
 
 static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
@@ -947,12 +915,16 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 	const uint8_t *pmkid;
 	struct ie_rsn_info rsn_info;
 
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
 	if (!eapol_verify_ptk_1_of_4(ek))
 		goto error_unspecified;
 
 	pmkid = handshake_util_find_pmkid_kde(ek->key_data,
 					L_BE16_TO_CPU(ek->key_data_len));
 
+	ie_parse_rsne_from_data(own_ie, own_ie[1] + 2, &rsn_info);
+
 	/*
 	 * Require the PMKID KDE whenever we've sent a list of PMKIDs in
 	 * our RSNE and we've haven't seen any EAPOL-EAP frame since
@@ -962,8 +934,6 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 	 * send no PMKID KDE.
 	 */
 	if (!sm->eap_exchanged && !sm->handshake->wpa_ie &&
-			ie_parse_rsne_from_data(own_ie, own_ie[1] + 2,
-						&rsn_info) >= 0 &&
 			rsn_info.num_pmkids) {
 		bool found = false;
 		int i;
@@ -984,6 +954,8 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 
 		if (handshake_state_get_pmkid(sm->handshake, own_pmkid) &&
 				memcmp(pmkid, own_pmkid, 16)) {
+			l_debug("Authenticator sent a PMKID that didn't match");
+
 			/*
 			 * If the AP has a different PMKSA from ours and we
 			 * have means to create a new PMKSA through EAP then
@@ -992,7 +964,18 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 			if (sm->eap) {
 				send_eapol_start(NULL, sm);
 				return;
-			} else
+			}
+
+			/*
+			 * Some APs are known to send a PMKID KDE with all
+			 * zeros for the PMKID.  Likely we can still
+			 * successfully negotiate a handshake, so ignore this
+			 * for now and treat it as if the PMKID KDE was not
+			 * included
+			 */
+			if (util_mem_is_zero(pmkid, 16))
+				l_debug("PMKID is all zero, ignoring");
+			else
 				goto error_unspecified;
 		}
 	}
@@ -1060,11 +1043,11 @@ static void eapol_handle_ptk_1_of_4(struct eapol_sm *sm,
 	}
 
 	memcpy(step2->key_mic_data, mic, sizeof(mic));
-	eapol_write(sm, (struct eapol_frame *) step2);
+	eapol_sm_write(sm, (struct eapol_frame *) step2, false);
 	l_free(step2);
 
-	l_timeout_remove(sm->timeout);
-	sm->timeout = NULL;
+	l_timeout_remove(sm->eapol_start_timeout);
+	sm->eapol_start_timeout = NULL;
 
 	return;
 
@@ -1072,6 +1055,144 @@ error_unspecified:
 	handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
 }
 
+#define EAPOL_PAIRWISE_UPDATE_COUNT 3
+
+/* 802.11-2016 Section 12.7.6.4 */
+static void eapol_send_ptk_3_of_4(struct eapol_sm *sm)
+{
+	uint32_t ifindex = sm->handshake->ifindex;
+	uint8_t frame_buf[512];
+	uint8_t key_data_buf[128];
+	struct eapol_key *ek = (struct eapol_key *) frame_buf;
+	size_t key_data_len;
+	enum crypto_cipher cipher = ie_rsn_cipher_suite_to_cipher(
+				sm->handshake->pairwise_cipher);
+	const struct crypto_ptk *ptk = (struct crypto_ptk *) sm->handshake->ptk;
+	struct ie_rsn_info rsn;
+
+	sm->replay_counter++;
+
+	memset(ek, 0, sizeof(struct eapol_key));
+	ek->header.protocol_version = sm->protocol_version;
+	ek->header.packet_type = 0x3;
+	ek->descriptor_type = EAPOL_DESCRIPTOR_TYPE_80211;
+	/* Must be HMAC-SHA1-128 + AES when using CCMP with PSK or 8021X */
+	ek->key_descriptor_version = EAPOL_KEY_DESCRIPTOR_VERSION_HMAC_SHA1_AES;
+	ek->key_type = true;
+	ek->install = true;
+	ek->key_ack = true;
+	ek->key_mic = true;
+	ek->secure = true;
+	ek->encrypted_key_data = true;
+	ek->key_length = L_CPU_TO_BE16(crypto_cipher_key_len(cipher));
+	ek->key_replay_counter = L_CPU_TO_BE64(sm->replay_counter);
+	memcpy(ek->key_nonce, sm->handshake->anonce, sizeof(ek->key_nonce));
+	/*
+	 * We don't currently handle group traffic, to support that we'd need
+	 * to provide the NL80211_ATTR_KEY_SEQ value from NL80211_CMD_GET_KEY
+	 * here.
+	 */
+	l_put_be64(1, ek->key_rsc);
+
+	/*
+	 * Just one RSNE in Key Data as we only set one cipher in ap->ciphers
+	 * currently.
+	 */
+
+	memset(&rsn, 0, sizeof(rsn));
+	rsn.akm_suites = IE_RSN_AKM_SUITE_PSK;
+	rsn.pairwise_ciphers = sm->handshake->pairwise_cipher;
+	rsn.group_cipher = IE_RSN_CIPHER_SUITE_NO_GROUP_TRAFFIC;
+
+	if (!ie_build_rsne(&rsn, key_data_buf))
+		return;
+
+	if (!eapol_encrypt_key_data(ptk->kek, key_data_buf,
+					2 + key_data_buf[1], ek))
+		return;
+
+	key_data_len = L_BE16_TO_CPU(ek->key_data_len);
+	ek->header.packet_len = L_CPU_TO_BE16(sizeof(struct eapol_key) +
+						key_data_len - 4);
+
+	if (!eapol_calculate_mic(ptk->kck, ek, ek->key_mic_data))
+		return;
+
+	l_debug("STA: "MAC" retries=%u", MAC_STR(sm->handshake->spa),
+			sm->frame_retry);
+
+	__eapol_tx_packet(ifindex, sm->handshake->spa, ETH_P_PAE,
+				(struct eapol_frame *) ek, false);
+}
+
+static void eapol_ptk_3_of_4_retry(struct l_timeout *timeout,
+						void *user_data)
+{
+	struct eapol_sm *sm = user_data;
+
+	if (sm->frame_retry >= EAPOL_PAIRWISE_UPDATE_COUNT) {
+		handshake_failed(sm, MMPDU_REASON_CODE_4WAY_HANDSHAKE_TIMEOUT);
+		return;
+	}
+
+	eapol_send_ptk_3_of_4(sm);
+
+	eapol_set_key_timeout(sm, eapol_ptk_3_of_4_retry);
+
+	l_debug("attempt %i", sm->frame_retry);
+}
+
+/* 802.11-2016 Section 12.7.6.3 */
+static void eapol_handle_ptk_2_of_4(struct eapol_sm *sm,
+					const struct eapol_key *ek)
+{
+	const uint8_t *rsne;
+	enum crypto_cipher cipher;
+	size_t ptk_size;
+	uint8_t ptk_buf[64];
+	struct crypto_ptk *ptk = (struct crypto_ptk *) ptk_buf;
+	const uint8_t *aa = sm->handshake->aa;
+
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
+	if (!eapol_verify_ptk_2_of_4(ek))
+		return;
+
+	if (L_BE64_TO_CPU(ek->key_replay_counter) != sm->replay_counter)
+		return;
+
+	cipher = ie_rsn_cipher_suite_to_cipher(sm->handshake->pairwise_cipher);
+	ptk_size = sizeof(struct crypto_ptk) + crypto_cipher_key_len(cipher);
+
+	if (!crypto_derive_pairwise_ptk(sm->handshake->pmk, sm->handshake->spa,
+					aa, sm->handshake->anonce,
+					ek->key_nonce, ptk, ptk_size, false))
+		return;
+
+	if (!eapol_verify_mic(ptk->kck, ek))
+		return;
+
+	/* Bitwise identical RSNE required */
+	rsne = eapol_find_rsne(ek->key_data,
+				L_BE16_TO_CPU(ek->key_data_len), NULL);
+	if (!rsne || rsne[1] != sm->handshake->own_ie[1] ||
+			memcmp(rsne + 2, sm->handshake->own_ie + 2, rsne[1])) {
+
+		handshake_failed(sm, MMPDU_REASON_CODE_IE_DIFFERENT);
+		return;
+	}
+
+	memcpy(sm->handshake->ptk, ptk_buf, ptk_size);
+	memcpy(sm->handshake->snonce, ek->key_nonce,
+			sizeof(sm->handshake->snonce));
+	sm->handshake->have_snonce = true;
+	sm->handshake->ptk_complete = true;
+
+	sm->frame_retry = 0;
+
+	eapol_ptk_3_of_4_retry(NULL, sm);
+}
+
 const uint8_t *eapol_find_rsne(const uint8_t *data, size_t data_len,
 				const uint8_t **optional)
 {
@@ -1133,6 +1254,8 @@ static void eapol_handle_ptk_3_of_4(struct eapol_sm *sm,
 	uint8_t gtk_key_index;
 	uint8_t igtk_key_index;
 
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
 	if (!eapol_verify_ptk_3_of_4(ek, sm->handshake->wpa_ie)) {
 		handshake_failed(sm, MMPDU_REASON_CODE_UNSPECIFIED);
 		return;
@@ -1328,7 +1451,7 @@ retransmit:
 	}
 
 	memcpy(step4->key_mic_data, mic, sizeof(mic));
-	eapol_write(sm, (struct eapol_frame *) step4);
+	eapol_sm_write(sm, (struct eapol_frame *) step4, false);
 	l_free(step4);
 
 	if (sm->handshake->ptk_complete)
@@ -1346,12 +1469,38 @@ retransmit:
 		rekey_offload(sm->handshake->ifindex, ptk->kek, ptk->kck,
 				sm->replay_counter, sm->user_data);
 
+	l_timeout_remove(sm->timeout);
+	sm->timeout = NULL;
+
 	return;
 
 error_ie_different:
 	handshake_failed(sm, MMPDU_REASON_CODE_IE_DIFFERENT);
 }
 
+/* 802.11-2016 Section 12.7.6.5 */
+static void eapol_handle_ptk_4_of_4(struct eapol_sm *sm,
+					const struct eapol_key *ek)
+{
+	const struct crypto_ptk *ptk = (struct crypto_ptk *) sm->handshake->ptk;
+
+	l_debug("ifindex=%u", sm->handshake->ifindex);
+
+	if (!eapol_verify_ptk_4_of_4(ek, false))
+		return;
+
+	if (L_BE64_TO_CPU(ek->key_replay_counter) != sm->replay_counter)
+		return;
+
+	if (!eapol_verify_mic(ptk->kck, ek))
+		return;
+
+	l_timeout_remove(sm->timeout);
+	sm->timeout = NULL;
+
+	handshake_state_install_ptk(sm->handshake);
+}
+
 static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 					const struct eapol_key *ek,
 					const uint8_t *decrypted_key_data,
@@ -1438,7 +1587,7 @@ static void eapol_handle_gtk_1_of_2(struct eapol_sm *sm,
 	}
 
 	memcpy(step2->key_mic_data, mic, sizeof(mic));
-	eapol_write(sm, (struct eapol_frame *) step2);
+	eapol_sm_write(sm, (struct eapol_frame *) step2, false);
 	l_free(step2);
 
 	eapol_install_gtk(sm, gtk_key_index, gtk, gtk_len, ek->key_rsc);
@@ -1595,7 +1744,7 @@ static void eapol_eap_msg_cb(const uint8_t *eap_data, size_t len,
 
 	memcpy(frame->data, eap_data, len);
 
-	eapol_write(sm, frame);
+	eapol_sm_write(sm, frame, false);
 }
 
 /* This respresentes the eapTimout, eapFail and eapSuccess messages */
@@ -1607,11 +1756,14 @@ static void eapol_eap_complete_cb(enum eap_result result, void *user_data)
 			"eapSuccess" : (result == EAP_RESULT_FAIL ?
 				"eapFail" : "eapTimeout"));
 
-	eap_free(sm->eap);
-	sm->eap = NULL;
-
-	if (result != EAP_RESULT_SUCCESS)
+	if (result != EAP_RESULT_SUCCESS) {
+		eap_free(sm->eap);
+		sm->eap = NULL;
 		handshake_failed(sm, MMPDU_REASON_CODE_IEEE8021X_FAILED);
+		return;
+	}
+
+	eap_reset(sm->eap);
 }
 
 /* This respresentes the eapResults message */
@@ -1621,8 +1773,6 @@ static void eapol_eap_results_cb(const uint8_t *msk_data, size_t msk_len,
 				void *user_data)
 {
 	struct eapol_sm *sm = user_data;
-	ssize_t pmk_len;
-	const uint8_t *pmk_data;
 
 	l_debug("EAP key material received");
 
@@ -1631,34 +1781,35 @@ static void eapol_eap_results_cb(const uint8_t *msk_data, size_t msk_len,
 	 *    "When not using a PSK, the PMK is derived from the AAA key.
 	 *    The PMK shall be computed as the first 256 bits (bits 0–255)
 	 *    of the AAA key: PMK ← L(PTK, 0, 256)."
-	 * 802.11 11.6.1.3:
+	 * 802.11-2016 12.7.1.3:
 	 *    "When not using a PSK, the PMK is derived from the MSK.
-	 *    The PMK shall be computed as the first 256 bits (bits 0–255)
-	 *    of the MSK: PMK ← L(MSK, 0, 256)."
+	 *    The PMK shall be computed as the first PMK_bits bits
+	 *    (bits 0 to PMK_bits–1) of the MSK: PMK = L(MSK, 0, PMK_bits)."
 	 * RFC5247 explains AAA-Key refers to the MSK and confirms the
 	 * first 32 bytes of the MSK are used.  MSK is at least 64 octets
 	 * long per RFC3748.  Note WEP derives the PTK from MSK differently.
 	 *
 	 * In a Fast Transition initial mobility domain association the PMK
-	 * maps to the XXKey except with EAP:
-	 * 802.11 11.6.1.7.3:
-	 *    "If the AKM negotiated is 00-0F-AC:3, then XXKey shall be the
-	 *    second 256 bits of the MSK (which is derived from the IEEE
+	 * maps to the XXKey, except with EAP:
+	 * 802.11-2016 12.7.1.7.3:
+	 *    "If the AKM negotiated is 00-0F-AC:3, then [...] XXKey shall be
+	 *    the second 256 bits of the MSK (which is derived from the IEEE
 	 *    802.1X authentication), i.e., XXKey = L(MSK, 256, 256)."
+	 * So we need to save the first 64 bytes at minimum.
 	 */
 
 	if (sm->handshake->akm_suite == IE_RSN_AKM_SUITE_FT_OVER_8021X) {
-		pmk_len = (ssize_t) msk_len - 32;
-		pmk_data = msk_data + 32;
+		if (msk_len < 64)
+			goto msk_short;
 	} else {
-		pmk_len = msk_len;
-		pmk_data = msk_data;
+		if (msk_len < 32)
+			goto msk_short;
 	}
 
-	if (pmk_len < 32)
-		goto msk_short;
+	if (msk_len > sizeof(sm->handshake->pmk))
+		msk_len = sizeof(sm->handshake->pmk);
 
-	handshake_state_set_pmk(sm->handshake, pmk_data);
+	handshake_state_set_pmk(sm->handshake, msk_data, msk_len);
 
 	return;
 
@@ -1693,6 +1844,54 @@ void eapol_sm_set_require_handshake(struct eapol_sm *sm, bool enabled)
 		sm->use_eapol_start = false;
 }
 
+static void eapol_auth_key_handle(struct eapol_sm *sm,
+				const struct eapol_frame *frame)
+{
+	size_t frame_len = 4 + L_BE16_TO_CPU(frame->header.packet_len);
+	const struct eapol_key *ek = eapol_key_validate((const void *) frame,
+							frame_len);
+
+	if (!ek)
+		return;
+
+	/* Wrong direction */
+	if (ek->key_ack)
+		return;
+
+	if (ek->request)
+		return; /* Not supported */
+
+	if (!sm->handshake->have_anonce)
+		return; /* Not expecting an EAPoL-Key yet */
+
+	if (!sm->handshake->ptk_complete)
+		eapol_handle_ptk_2_of_4(sm, ek);
+	else
+		eapol_handle_ptk_4_of_4(sm, ek);
+}
+
+static void eapol_rx_auth_packet(uint16_t proto, const uint8_t *from,
+				const struct eapol_frame *frame,
+				void *user_data)
+{
+	struct eapol_sm *sm = user_data;
+
+	if (!sm->protocol_version)
+		sm->protocol_version = frame->header.protocol_version;
+
+	switch (frame->header.packet_type) {
+	case 3: /* EAPOL-Key */
+		eapol_auth_key_handle(sm, frame);
+		break;
+
+	default:
+		l_error("Authenticator received unknown packet type %i from %s",
+			frame->header.packet_type,
+			util_address_to_string(from));
+		return;
+	}
+}
+
 static void eapol_rx_packet(uint16_t proto, const uint8_t *from,
 				const struct eapol_frame *frame,
 				void *user_data)
@@ -1796,11 +1995,6 @@ void __eapol_set_tx_user_data(void *user_data)
 	tx_user_data = user_data;
 }
 
-void __eapol_set_deauthenticate_func(eapol_deauthenticate_func_t func)
-{
-	deauthenticate = func;
-}
-
 void __eapol_set_rekey_offload_func(eapol_rekey_offload_func_t func)
 {
 	rekey_offload = func;
@@ -1814,6 +2008,20 @@ void eapol_register(struct eapol_sm *sm)
 						eapol_rx_packet, sm);
 }
 
+void eapol_register_authenticator(struct eapol_sm *sm)
+{
+	l_queue_push_head(state_machines, sm);
+
+	sm->watch_id = eapol_frame_watch_add(sm->handshake->ifindex,
+			eapol_rx_auth_packet, sm);
+
+	sm->started = true;
+	sm->authenticator = true;
+
+	/* kick off handshake */
+	eapol_ptk_1_of_4_retry(NULL, sm);
+}
+
 bool eapol_start(struct eapol_sm *sm)
 {
 	if (sm->handshake->settings_8021x) {
@@ -1837,7 +2045,8 @@ bool eapol_start(struct eapol_sm *sm)
 	sm->started = true;
 
 	if (sm->require_handshake)
-		sm->timeout = l_timeout_create(2, eapol_timeout, sm, NULL);
+		sm->timeout = l_timeout_create(eapol_4way_handshake_time,
+				eapol_timeout, sm, NULL);
 
 	if (sm->use_eapol_start) {
 		/*
@@ -1900,12 +2109,6 @@ bool eapol_frame_watch_remove(uint32_t id)
 	return watchlist_remove(&frame_watches, id);
 }
 
-void eapol_tx_frame(uint32_t ifindex, uint16_t proto, const uint8_t *dst,
-			const struct eapol_frame *frame)
-{
-	pae_write(ifindex, dst, NULL, proto, frame);
-}
-
 struct preauth_sm {
 	uint32_t ifindex;
 	uint8_t aa[6];
@@ -1948,7 +2151,7 @@ static void preauth_frame(struct preauth_sm *sm, uint8_t packet_type,
 	if (data_len)
 		memcpy(frame->data, data, data_len);
 
-	pae_write(sm->ifindex, sm->aa, sm->spa, 0x88c7, frame);
+	__eapol_tx_packet(sm->ifindex, sm->aa, 0x88c7, frame, false);
 }
 
 static void preauth_rx_packet(uint16_t proto, const uint8_t *from,
@@ -2119,7 +2322,8 @@ static bool eapol_frame_watch_match_ifindex(const void *a, const void *b)
 }
 
 void __eapol_rx_packet(uint32_t ifindex, const uint8_t *src, uint16_t proto,
-					const uint8_t *frame, size_t len)
+					const uint8_t *frame, size_t len,
+					bool noencrypt)
 {
 	const struct eapol_header *eh;
 
@@ -2147,6 +2351,22 @@ void __eapol_rx_packet(uint32_t ifindex, const uint8_t *src, uint16_t proto,
 					(const struct eapol_frame *) eh);
 }
 
+void __eapol_tx_packet(uint32_t ifindex, const uint8_t *dst, uint16_t proto,
+			const struct eapol_frame *frame, bool noencrypt)
+{
+	if (!tx_packet)
+		return;
+
+	tx_packet(ifindex, dst, proto, frame, noencrypt, tx_user_data);
+}
+
+void __eapol_set_config(struct l_settings *config)
+{
+	if (!l_settings_get_uint(config, "EAPoL",
+			"max_4way_handshake_time", &eapol_4way_handshake_time))
+		eapol_4way_handshake_time = 2;
+}
+
 bool eapol_init()
 {
 	state_machines = l_queue_new();
